"""
env: python 3.8

This lambda task is triggered by a new S3 event published to the SQS (see trigger module for the triggering queue).

It takes one message from the queue and extracts the newly uploaded object information, including S3 bucket name and
S3 object key. Using these information, it calls Rekognition's PPE detection API and receives the model inference
as a JSON response. This function then further calculates PPE detection work stream metrics and writes the output
to Kinesis Firehose to be delivered back to data lake.

"""

# requirements
import os
import json
import boto3
import logging
from collections import Counter

# initiating logger
logger = logging.getLogger()
logger.setLevel(logging.INFO)

s3_client = boto3.resource('s3', region_name=os.environ['AWS_REGION'])
rekognition = boto3.client("rekognition", region_name=os.environ['AWS_REGION'])
firehose_client = boto3.client('firehose', region_name=os.environ['AWS_REGION'])

def lambda_handler(event, context):
    """ main lambda entry point (invoke point)

    This Lambda receives a message from SQS, extracts newly uploaded image object information and submits them
    to Rekognition API to get PPE detection JSON response. After receiving the response, it calculates metrics
    to report and send them to Firehose to be delivered.

    @param event: SQS event
    @param context: Lambda context information

    @return None

    """
    logger.info('Triggered event: {}'.format(json.dumps(event)))

    # load the new event from SQS
    put_event = json.loads(event['Records'][0]['body'])

    # check if it's a s3 test event
    if 'Event' in put_event.keys():
        logger.info('S3 test event identified. No action is taken.')
        event_type = put_event['Event']
        if event_type == 's3:TestEvent':
            return {'statusCode': 200}

    # Get the object bucket and name (key) from event
    bucket = put_event['Records'][0]['s3']['bucket']['name']
    key = put_event['Records'][0]['s3']['object']['key']

    logger.info('Bucket: {}'.format(bucket))
    logger.info('Object Key: {}'.format(key))

    # # extract camera ID and object (image) information
    _metadata = key.split('/')
    camera_id = _metadata[1]

    try:
        # Calling Rekognition API for PPE detection
        ppe_response = get_ppe_response(bucket, key)

        # calculate metrics
        logger.info('Calculating metrics.')
        agg_res = Counter(ppe_response.values())
        person_total = agg_res['non-compliant'] + agg_res['compliant'] + agg_res['unknown']
        detected_total = agg_res['non-compliant'] + agg_res['compliant']
        face_cover_noncompliant = agg_res['non-compliant']

        if detected_total > 0:
            face_cover_noncompliant_percentage = round(agg_res['non-compliant'] * 100. / detected_total, 2)
        else:
            face_cover_noncompliant_percentage = 0

        if person_total > 0:
            detected_face_rate = round(detected_total * 100. / person_total, 2)
        else:
            detected_face_rate = 0

        # compose Firehose message
        message = {
            "CameraID": camera_id,
            "Bucket": bucket,
            "Key": key,
            "RekResult": json.dumps(ppe_response),
            "FaceCoverNonCompliant": face_cover_noncompliant,
            "FaceCoverNonCompliantPercentage": face_cover_noncompliant_percentage,
            "detectedFaceRate": detected_face_rate
        }

        # write to Kinesis Firehose
        success = output_to_firehose(message)

        if success:
            logger.info('PPE detection job completed.')
            return {'statusCode': 200}
        else:
            logger.info('Failed to send PPE detection result to Firehose.')
            return {'statusCode': 400}

    except Exception as e:
        logger.info('Rekognition job submission failed with error: {}'.format(e))

        # write out an error record
        s3_client.Bucket(bucket).put_object(
            Body=json.dumps({'bucket': bucket, 'key': key, 'error': str(e)}),
            Key="error/ppe_detection/failed-job.json"
        )

        raise e


def get_ppe_response(bucket, key):
    """ calls Rekognition API, determines if people detected are wearing PPE properly, and outputs
    the predicted outcome as a dictionary

    @param bucket: S3 bucket name
    @param key: S3 object key

    @return ppe_response: a dictionary where the key is person ID and value is whether the person is non-compliant
     with PPE guidelines (i.e. does the person wears PPE properly)
    """
    logger.info('Calling Rekognition API for PPE detection.')
    response = rekognition.detect_protective_equipment(
        Image={"S3Object": {"Bucket": bucket, "Name": key}}
    )

    if response['ResponseMetadata']['HTTPStatusCode'] == 200:
        logger.info('Rekognition job submission succeeded.')
    else:
        logger.info('Rekognition job submission failed.')
        return False

    # Append Person_ID and noncompliant in two separate lists for each person detected in frame
    person_id = []
    ppe_noncompliant = []

    num_person_detected = len(response['Persons'])

    if num_person_detected == 0:
        return {}

    logger.info('Rekognition response: {}'.format(json.dumps(response)))
    for i in range(num_person_detected):
        instance = response['Persons'][i]
        if instance['Confidence'] >= 90:
            bodyparts = instance['BodyParts']
            if len(bodyparts) == 0:
                continue
            l1 = instance['BodyParts'][0]['EquipmentDetections']
            noncompliant = is_person_ppe_noncompliant(l1)
            ppe_noncompliant.append(noncompliant)
            p_id = instance['Id']
            person_id.append(p_id)

    ppe_response = dict(zip(person_id, ppe_noncompliant))

    return ppe_response


def is_person_ppe_noncompliant(ppe_list):
    """ Checks if the face is protected with mask

    @param ppe_list: takes the response list

    @return "non-compliant": Person is non-compliant PPE
    @return "compliant": Person is wearing mask properly
    @return "unknown": Unable to decide
    """
    if no_face_detected(ppe_list) == 0:
        if ppe_list[0]['CoversBodyPart']['Value'] is True:
            return "compliant"
        return "non-compliant"
    else:
        return "unknown"


def no_face_detected(ppe_list):
    """ Checks if the face is detected.

    @param ppe_list: takes the response list

    @return 0: if face is detected
    @return 1: if face is not detected
    """
    if not ppe_list:
        return 1
    else:
        return 0


def output_to_firehose(message):
    """ Takes in the Firehose message and sends it through Firehose client

    @param message: dictionary containing Firehose message

    @return True: if message is not sent successfully
    @return False: if message is not sent successfully
    """
    try:
        # get fire hose delivery stream name
        firehose_delivery_stream = os.environ['firehose_name']

        # if it's an arn, extract only the delivery stream name
        if firehose_delivery_stream.startswith('arn'):
            firehose_delivery_stream = firehose_delivery_stream.split('deliverystream/')[-1]

        # sends the message through Firehose client
        response = firehose_client.put_record(
            DeliveryStreamName=firehose_delivery_stream,
            Record={'Data': json.dumps(message)+'\n'}
        )

        logger.info('Message sent: {}'.format(json.dumps(message)))
        logger.info('Firehose response: {}'.format(json.dumps(response)))

    except Exception as e:

        logger.info('Failed to send message to Firehose')
        logger.info('Error: {}'.format(e))

        return False

    return True